
SERVICE_SUMMARY_MOCK = """
{
    "documentId": "683",
    "services": [
        {
            "c": "101",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "sd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One Lon",
            "xsg": 3
        },
        {
            "c" : "811",
            "dvbtriplet" : "2.2022.6506",
            "schedule" : true,
            "servicetype" : "DSAT",
            "sf" : "sd",
            "sg" : 12,
            "sid" : "2306",
            "sk" : 2306,
            "t" : "Dave",
            "xsg" : 3
        }
    ],
    "version": 3
}
"""

SERVICE_DETAIL_1 = """
{
    "details": {
        "dvbtriplet": "2.2045.6301",
        "isbroadcasting": true,
        "upgradeMessage": "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."
    },
    "sid": "2002",
    "streamingprofiles": [
        {
            "name": "AnExample",
            "suri": "http://10.0.1.6:4730/trans_caption/CHAN%3Alocator%3A5%3A3%3A7D2/profileAnExample.ttml",
            "uri": "http://10.0.1.6:4730/trans/CHAN%3Alocator%3A5%3A3%3A7D2/profileAnExample.ts"
        }
    ]
}
"""

SERVICE_DETAIL_2 = """
{
    "details": {
        "dvbtriplet": "2.2022.6506",
        "isbroadcasting": true,
        "upgradeMessage": "Dave is the home of witty banter with quizcoms, cars and comedies."
    },
    "sid": "2306",
    "streamingprofiles": [
        {
            "name": "AnExample",
            "suri": "http://10.0.1.6:4730/trans_caption/CHAN%3Alocator%3A5%3A3%3A902/profileAnExample.ttml",
            "uri": "http://10.0.1.6:4730/trans/CHAN%3Alocator%3A5%3A3%3A902/profileAnExample.ts"
        }
    ]
}
"""

REMOTE_TCP_MOCK = [  # set up the data to be returned on each successive call of socket.recv()
    b'SKY 000.001\n',
    b'\x01\x01',
    b'\x00\x00\x00\x00',
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
]


WS_STATUS_MOCK = """
{
   "camessage" : {
      "reason" : "no message",
      "state" : "unavailable"
   },
   "drmstatus" : {
      "state" : "available"
   },
   "entitlements" : [
      "ANALYTICS",
      "BIGBASIC",
      "ETHAN_APP_1",
      "HD",
      "PDL",
      "SKY_DRM_CE",
      "SKY_DRM_MR",
      "SKY_IPPV",
      "ULTRA+",
      "SKY+",
      "GATEWAYENABLER",
      "SIDELOAD"
   ],
   "epginfobits" : {
      "epginfobits" : "0xFDE5FFC0",
      "mask" : "0x5FFA003F",
      "state" : "available"
   },
   "gatewayservices" : {
      "state" : "available"
   },
   "hdmi" : {
      "2160p10bitCapable" : false,
      "authenticatedHDCP" : "NONE",
      "reason" : "HDMI output port is disabled",
      "sinkHDCP" : "NONE",
      "sinkHLG" : false,
      "sinkUHD" : false,
      "state" : "available",
      "uhdConfigured" : false
   },
   "network" : {
      "state" : "available"
   },
   "nssplayback" : {
      "state" : "available"
   },
   "pvr" : {
      "state" : "available"
   },
   "schedule" : {
      "lastdate" : "20181005",
      "state" : "available"
   },
   "servicelist" : {
      "state" : "available"
   },
   "smartcard" : {
      "active" : true,
      "bouquet" : "4101",
      "countryCode" : "GBR",
      "currency" : "GBP",
      "cwe" : true,
      "householdid" : "12345678",
      "paired" : true,
      "state" : "available",
      "subbouquet" : "1",
      "transactionlimit" : 65535,
      "viewingCardNumber" : "123 456 789"
   },
   "swupdate" : {
      "reason" : "IDLE",
      "state" : "unavailable"
   },
   "systemupdates" : {
      "entitlements" : 2,
      "install" : 1,
      "servicegenres" : 1,
      "smartcard" : 1
   },
   "updatetask" : {
      "reason" : "no update",
      "state" : "unavailable"
   }
}

"""

XML_CHANNEL_1 = """
  <channel id="f3932e75f691561adbe3b609369e487b">
    <display-name>BBC One Lon</display-name>
    <blah>Not understood, not processed, so should be ignored.</blah>
    <icon src="/images/channels/f3932e75f691561adbe3b609369e487b.png"/>
  </channel>
"""

XML_CHAN_JSON = """{
    "__type__": "__channel__",
    "attributes": {
        "sid": null,
        "c": null,
        "t": null,
        "xmltv_id": "f3932e75f691561adbe3b609369e487b",
        "xmltv_display_name": "BBC One Lon",
        "xmltv_icon_url": "http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png"
    },
    "sources": 4,
    "programmes": []
}
"""

SKYQ_CHAN_JSON = """{
    "__type__": "__channel__",
    "attributes": {
        "sid": "2002",
        "c": "101",
        "t": "BBC One Lon",
        "xmltv_id": null,
        "dvbtriplet": "2.2045.6301",
        "schedule": "true",
        "servicetype": "DSAT",
        "sf": "sd",
        "sg": 12,
        "sk": 2002,
        "xsg": 3,
        "isbroadcasting": true,
        "upgradeMessage": "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."
    },
    "sources": 3,
    "programmes": []
}
"""

MERGED_CHAN_JSON = """{
    "__type__": "__channel__",
    "attributes": {
        "sid": "2002",
        "c": "101",
        "t": "BBC One Lon",
        "xmltv_id": "f3932e75f691561adbe3b609369e487b",
        "dvbtriplet": "2.2045.6301",
        "schedule": "true",
        "servicetype": "DSAT",
        "sf": "sd",
        "sg": 12,
        "sk": 2002,
        "xsg": 3,
        "isbroadcasting": true,
        "upgradeMessage": "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england.",
        "xmltv_display_name": "BBC One Lon",
        "xmltv_icon_url": "http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png"
    },
    "sources": 7,
    "programmes": []
}
"""


BAD_JSON = """
{
    "attributes": {
        "sid": "2002",
        "c": "101",
        "t": "BBC One Lon",
        "xmltv_id": "f3932e75f691561adbe3b609369e487b",
        "dvbtriplet": "2.2045.6301",
        "schedule": "true",
        "servicetype": "DSAT",
        "sf": "sd",
        "sg": 12,
        "sk": 2002,
        "xsg": 3,
        "isbroadcasting": true,
        "upgradeMessage": "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england.",
        "xmltv_display_name": "BBC One Lon",
        "xmltv_icon_url": "http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png"
    },
    "sources": 7
}
"""
FUZZY_CHANNELS_MOCK = """
{
    "documentId": "683",
    "services": [
        {
            "c": "101",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "sd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One Lon",
            "xsg": 3
        },
        {
            "c": "102",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "hd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One Lon HD",
            "xsg": 3
        },
        {
            "c": "103",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "hd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One HD",
            "xsg": 3
        },
        {
            "c": "104",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "sd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One +1",
            "timeshifted": true,
            "xsg": 3
        },
        {
            "c": "105",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "sd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC News",
            "xsg": 3
        },
        {
            "c": "106",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "sd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One YH",
            "xsg": 3
        },
        {
            "c": "107",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "hd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One NI HD",
            "xsg": 3
        },
        {
            "c": "108",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "sd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC News Lon",
            "xsg": 3
        },
        {
            "c": "109",
            "dvbtriplet": "2.2045.6301",
            "schedule": "true",
            "servicetype": "DSAT",
            "sf": "hd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC News HD",
            "xsg": 3
        },
        {
            "c" : "110",
            "dvbtriplet" : "2.2022.6506",
            "schedule" : true,
            "servicetype" : "DSAT",
            "sf" : "hd",
            "sg" : 12,
            "sid" : "2306",
            "sk" : 2306,
            "t" : "BBC News HD +1",
            "timeshifted": true,
            "xsg" : 3
        }
    ],
    "version": 3
}
"""
# EPG_JSON = """
# [
#     {
#         "__type__": "__channel__",
#         "attributes": {
#             "sid": "2002",
#             "c": "101",
#             "t": "BBC One Lon",
#             "xmltv_id": null,
#             "dvbtriplet": "2.2045.6301",
#             "schedule": "true",
#             "servicetype": "DSAT",
#             "sf": "sd",
#             "sg": 12,
#             "sk": 2002,
#             "xsg": 3,
#             "isbroadcasting": true,
#             "upgradeMessage": "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."
#         },
#         "sources": 3,
#         "programmes": []

#     },
#     {
#         "__type__": "__channel__",
#         "attributes": {
#             "sid": "2306",
#             "c": "811",
#             "t": "Dave",
#             "xmltv_id": null,
#             "dvbtriplet": "2.2022.6506",
#             "schedule": true,
#             "servicetype": "DSAT",
#             "sf": "sd",
#             "sg": 12,
#             "sk": 2306,
#             "xsg": 3,
#             "isbroadcasting": true,
#             "upgradeMessage": "Dave is the home of witty banter with quizcoms, cars and comedies."
#         },
#         "sources": 3,
#         "programmes": []
#     },
#     {
#         "__type__": "__SOMETHING_TBC__",
#         "attributes": {
#             "blah": "yadda"
#         }
#     }
# ]
# """



"""This module implements the Programme class."""

from datetime import datetime
from typing import Optional
from xml.etree.ElementTree import Element

from dateutil.parser import parse

from dataclasses import dataclass, field


# This object needs a few special attributes:
# - The ability to be hashable so that it can reside in a Sorted set
#   reflecting the programs in order on a channel.
# - The ability to support greater than, less than or equal to comparisons using
#   ony the start time.
#
# We should bear in mind that the detail of a particular programme as provided
# by the XMLTV file might improve in quality as XMLTV files come in various
# flavours (e.g., long forward range but limited data, or shorter forward
# ranges, but with richer data). We want to be able to update a Programme with
# richer data as its show time nears without making it appear as if it's a new
# programme, so we want its hash to remain the same as long as the Title,
# Channel and Start Time are the same.
#
# A Sorted Set maintains both deduplication and ordering simultaneously and so
# it should be a good structure to allow us to add newer entries of the same
# programme as the data gets richer, while preserving order via the programme's
# hash.


@dataclass(order=True)  # pylint: disable=too-few-public-methods
class Programme:
    """This dataclass represents a TV Programme.

    Attributes:
        title (str): Title of the Programme.
        title_lang (str): Language that the Title is in.
        desc (str): Description of the Programme.
        desc_lang (str): Language that the Description is in.
        start (datetime): Start datetime.
        start_raw (str): Start datetime in raw string format.
        stop (datetime): End datetime.
        stop_raw (str):  End datetime in raw string format.
        channel_xml_id (str): Channel key. Maps to :attr:`pyskyq.channel.Channel.xmltv_id`
        episode_num (str): Optional series and episode information.
        episode_num_system (str): Numbering scheme used for episode data.

    """

    title: str = field(init=True,
                       repr=True,
                       compare=False,
                       )
    title_lang: str = field(init=True,
                            repr=False,
                            compare=False,
                            )
    desc: str = field(init=True,
                      repr=False,
                      compare=False,
                      )
    desc_lang: str = field(init=True,
                           repr=False,
                           compare=False,
                           )
    start: datetime = field(init=False,  # computed value
                            repr=True,
                            compare=True,
                            )
    start_raw: str = field(init=True,
                           repr=False,
                           compare=False,
                           )
    stop: datetime = field(init=False,  # computed value
                           repr=True,
                           compare=False,
                           )
    stop_raw: str = field(init=True,
                          repr=False,
                          compare=False,
                          )
    channel_xml_id: str = field(init=True,
                                repr=True,
                                compare=False,
                                )
    episode_num: Optional[str] = field(init=True,
                                       repr=True,
                                       compare=False,
                                       )
    episode_num_system: Optional[str] = field(init=True,
                                              repr=False,
                                              compare=False,
                                              )

    def __post_init__(self):
        """Process the start and stop strings into datetime objects."""
        self.start = parse(self.start_raw)
        self.stop = parse(self.stop_raw)

    def __hash__(self):
        """Calculate the hash of this object."""
        return hash(self.title + self.channel_xml_id + str(self.start))

def programme_from_xmltv_list(xml_prog: Element) -> Programme:
    """Create a Programme object from an XMLTV programme element.

    This function is a Programme factory. It will generate a Programme object
    given a XML channel element from an XML TV file.

    Args:
        xml_prog (Element): A XML element of ``<programme>...</programme>`` tags.

    Returns:
        Programme: A channel object with the XML TV data loaded.

    """
    assert xml_prog.tag.lower() == 'programme'

    start = xml_prog.attrib['start']
    stop = xml_prog.attrib['stop']
    chan_id = xml_prog.attrib['channel']
    episode_num_system = None
    episode_num = None

    for child in xml_prog:
        if child.tag.lower() == 'title':
            title_lang = child.attrib.get('lang', None)
            title = child.text
            continue
        if child.tag.lower() == 'desc':
            desc_lang = child.attrib.get('lang', None)
            desc = child.text
            continue
        if child.tag.lower() == 'episode-num':
            episode_num_system = child.attrib.get('system', None)
            episode_num = child.text
            continue

    return Programme(title=title,     # type: ignore
                     title_lang=title_lang,
                     desc=desc,
                     desc_lang=desc_lang,
                     start_raw=start,
                     stop_raw=stop,
                     channel_xml_id=chan_id,
                     episode_num=episode_num,
                     episode_num_system=episode_num_system
                     )



# pylint: skip-file
import json
import xml.etree.ElementTree as ET

import pytest
from yarl import URL

from pyskyq import (CHANNELSOURCES, Channel, channel_from_skyq_service,
                    channel_from_xmltv_list, merge_channels, channel_from_json)

from .mock_constants import (MERGED_CHAN_JSON, SERVICE_DETAIL_1,
                             SERVICE_SUMMARY_MOCK, SKYQ_CHAN_JSON,
                             XML_CHAN_JSON, XML_CHANNEL_1, BAD_JSON)


def test_hashable_channels():

    blank_chan = Channel.__new__(Channel)
    another_blank_chan = Channel.__new__(Channel)

    # all blank channels are the same
    assert hash(blank_chan) == hash(another_blank_chan)

    skyq_chan1 = channel_from_skyq_service(json.loads(SERVICE_SUMMARY_MOCK)['services'][0])
    skyq_chan2 = channel_from_skyq_service(json.loads(SERVICE_SUMMARY_MOCK)['services'][0])

    # all sky channels from the same source are the same...
    assert hash(skyq_chan1) == hash(skyq_chan2)
    # but different to blank ones
    assert hash(blank_chan) != hash(skyq_chan1)

    skyq_chan1.load_skyq_detail_data(json.loads(SERVICE_DETAIL_1))

    # a sky channel with detail is the same as one with only summary data.
    assert hash(skyq_chan1) == hash(skyq_chan2)

    xml_chan = channel_from_xmltv_list(ET.XML(XML_CHANNEL_1))

    # channels with the same name, but diffenent sources are the same.
    assert skyq_chan1.name == xml_chan.xmltv_display_name
    assert hash(skyq_chan1) == hash(xml_chan)
    assert hash(skyq_chan1) == hash(xml_chan)

    # a merging of 2 channels with the same name is the same as both of the sources.
    assert hash(merge_channels(skyq_chan1, xml_chan)) == \
        hash(skyq_chan1)

    assert hash(merge_channels(skyq_chan1, xml_chan)) == \
        hash(xml_chan)



def test_blank_channel():

    with pytest.raises(NotImplementedError,
                       match='This class cannot be instantiated directly. ' +
                       'Use a factory function to create an instance.'
                       ):
        blank_chan = Channel()

    blank_chan = Channel.__new__(Channel)


    assert blank_chan.sources == CHANNELSOURCES.no_source
    assert blank_chan.__repr__() == f'<Channel: sources={CHANNELSOURCES.no_source}, id=None, xmltv_id=None, number=None, name=None>'

    with pytest.raises(AttributeError, match="Can't modify sid"):
        blank_chan.sid = 232

    assert blank_chan.blah is None

    assert blank_chan.isbroadcasting is None

def test_channel_from_skyq_service():

    chan = channel_from_skyq_service(json.loads(SERVICE_SUMMARY_MOCK)['services'][0])

    assert isinstance(chan, Channel)
    assert chan.sources == CHANNELSOURCES.skyq_service_summary
    assert chan.__repr__() == f'<Channel: sources={CHANNELSOURCES.skyq_service_summary}, id=2002, xmltv_id=None, number=101, name=BBC One Lon>'
    assert chan.c == "101"
    assert chan.t == "BBC One Lon"
    assert chan.name == "BBC One Lon"

    with pytest.raises(AttributeError, match="Can't modify sid"):
        chan.sid = 232

    assert chan.isbroadcasting is None

    chan.load_skyq_detail_data(json.loads(SERVICE_DETAIL_1))

    assert chan.isbroadcasting
    assert chan.sources == CHANNELSOURCES.skyq_service_summary | CHANNELSOURCES.skyq_service_detail
    assert chan.__repr__() == f'<Channel: sources={CHANNELSOURCES.skyq_service_detail | CHANNELSOURCES.skyq_service_summary}, id=2002, xmltv_id=None, number=101, name=BBC One Lon>'

    assert chan.upgradeMessage == "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."
    assert chan.desc == "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."


def test_channel_from_xmltv_data():

    chan = channel_from_xmltv_list(ET.XML(XML_CHANNEL_1))

    assert isinstance(chan, Channel)
    assert chan.sources == CHANNELSOURCES.xml_tv
    assert chan.__repr__() == f'<Channel: sources={CHANNELSOURCES.xml_tv}, id=None, xmltv_id=f3932e75f691561adbe3b609369e487b, number=None, name=None>'
    assert chan.xmltv_id == "f3932e75f691561adbe3b609369e487b"
    assert isinstance(chan.xmltv_icon_url, URL)
    assert chan.xmltv_icon_url.human_repr() == "http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png"
    assert chan.xmltv_display_name == "BBC One Lon"


def test_channel_from_both_sources():
    chan_x = channel_from_xmltv_list(ET.XML(XML_CHANNEL_1))
    chan_q = channel_from_skyq_service(json.loads(SERVICE_SUMMARY_MOCK)['services'][0])

    # first way round
    chan = merge_channels(chan_q, chan_x) # note, chan_x overwrites chan_q

    assert isinstance(chan, Channel)
    assert chan.sources == CHANNELSOURCES.xml_tv | CHANNELSOURCES.skyq_service_summary
    assert chan.__repr__() == f'<Channel: sources={CHANNELSOURCES.xml_tv | CHANNELSOURCES.skyq_service_summary}, id=2002, xmltv_id=f3932e75f691561adbe3b609369e487b, number=101, name=BBC One Lon>'
    assert chan.c == "101"
    assert chan.t == "BBC One Lon"
    assert chan.name == "BBC One Lon"
    assert chan.xmltv_id == "f3932e75f691561adbe3b609369e487b"
    assert isinstance(chan.xmltv_icon_url, URL)
    assert chan.xmltv_icon_url.human_repr() == "http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png"
    assert chan.xmltv_display_name == "BBC One Lon"

    # other way round
    chan = merge_channels(chan_x, chan_q)  # note, chan_q overwrites chan_x

    assert isinstance(chan, Channel)
    assert chan.sources == CHANNELSOURCES.xml_tv | CHANNELSOURCES.skyq_service_summary
    assert chan.__repr__() == f'<Channel: sources={CHANNELSOURCES.xml_tv | CHANNELSOURCES.skyq_service_summary}, id=2002, xmltv_id=f3932e75f691561adbe3b609369e487b, number=101, name=BBC One Lon>'
    assert chan.c == "101"
    assert chan.t == "BBC One Lon"
    assert chan.name == "BBC One Lon"
    assert chan.xmltv_id == "f3932e75f691561adbe3b609369e487b"
    assert isinstance(chan.xmltv_icon_url, URL)
    assert chan.xmltv_icon_url.human_repr() == "http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png"
    assert chan.xmltv_display_name == "BBC One Lon"


def test_channel_json():
    chan_x = channel_from_xmltv_list(ET.XML(XML_CHANNEL_1))
    chan_q = channel_from_skyq_service(json.loads(SERVICE_SUMMARY_MOCK)['services'][0])
    chan_q.load_skyq_detail_data(json.loads(SERVICE_DETAIL_1))

    mergechan = merge_channels(chan_q, chan_x)  # note, chan_x overwrites chan_q

    assert json.loads(chan_x.as_json()) == json.loads(XML_CHAN_JSON)
    assert json.loads(chan_q.as_json()) == json.loads(SKYQ_CHAN_JSON)
    assert json.loads(mergechan.as_json()) == json.loads(MERGED_CHAN_JSON)

    assert chan_x == channel_from_json(XML_CHAN_JSON)
    assert chan_q == channel_from_json(SKYQ_CHAN_JSON)
    assert mergechan == channel_from_json(MERGED_CHAN_JSON)

    assert chan_x != chan_q

    assert chan_x != "blah"

    with pytest.raises(ValueError, match="Incorrect type metadata in JSON payload."):
        channel_from_json(BAD_JSON)





import logging
import sys
from functools import partial
from pathlib import Path
from dateutil.parser import parse

import pytest
import trio

from pyskyq import EPG, XMLTVListing, Programme

from .http_server import http_server
from .mock_constants import (SERVICE_DETAIL_1, SERVICE_DETAIL_2,
                             SERVICE_SUMMARY_MOCK)

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.WARNING, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

LOGGER = logging.getLogger(__name__)


json_data = ""


async def test_EPG_sky_channels():

    epg = EPG('localhost', rest_port=8000)

    with pytest.raises(ValueError, match='No channels loaded.'):
        epg.get_channel_by_sid(2002)

    with trio.move_on_after(5):
        async with trio.open_nursery() as server_nursery:
            LOGGER.debug('in nursery')
            responses = [
                {
                    'target': '/as/services',
                    'status_code': 200,
                    'content_type': "application/json; charset=utf-8",
                    'body': SERVICE_SUMMARY_MOCK.encode('utf-8'),
                    'headers': {
                        'Connection': 'close'
                    },
                },
                {
                    'target': '/as/services/details/2002',
                    'status_code': 200,
                    'content_type': "application/json; charset=utf-8",
                    'body': SERVICE_DETAIL_1.encode('utf-8'),
                    'headers': {
                        'Connection': 'close'
                    },
                },
                {
                    'target': '/as/services/details/2306',
                    'status_code': 200,
                    'content_type': "application/json; charset=utf-8",
                    'body': SERVICE_DETAIL_2.encode('utf-8'),
                    'headers': {
                        'Connection': 'close'
                    },
                },
            ]

            await server_nursery.start(trio.serve_tcp, partial(http_server, responses=responses), 8000)
            # await trio.serve_tcp(partial(http_server, responses=responses), 8000)
            LOGGER.debug('started server.')
            await epg.load_skyq_channel_data()

    assert isinstance(epg, EPG)
    assert len(epg._channels) == 2
    assert epg.get_channel_by_sid(2002).c == "101"
    assert epg.get_channel_by_sid(2002).t == "BBC One Lon"
    assert epg.get_channel_by_sid(2002).name == "BBC One Lon"
    assert epg.get_channel_by_sid('2002').c == "101"
    assert epg.get_channel_by_sid('2002').t == "BBC One Lon"
    assert epg.get_channel_by_sid('2002').name == "BBC One Lon"

    assert epg.get_channel_by_sid(2002).isbroadcasting is True
    assert "BBC ONE for Greater London and the surrounding area." in \
        epg.get_channel_by_sid(2002).upgradeMessage

    assert "Dave is the home of witty banter with quizcoms, cars and comedies." in \
        epg.get_channel_by_sid(2306).upgradeMessage

    with pytest.raises(ValueError, match='Sid:1234567 not found.'):
        epg.get_channel_by_sid(1234567)

    global json_data
    json_data = epg.as_json()   # save state here so we can use in the next test.


xmlfile_path = Path(__file__).resolve().parent.joinpath('parse_xmltv_data.xml')


def test_apply_EPG_XMLTV_listing():

    epg = EPG('localhost', rest_port=8000)

    global json_data

    epg.from_json(json_data)

    l = XMLTVListing('http://host.com/some/feed')
    l._full_path = xmlfile_path

    with pytest.raises(ValueError, match='No XMLTVListing file found.'):
        epg.apply_XMLTVListing(l)

    l._downloaded = True

    # now test the xml loading.
    epg.apply_XMLTVListing(l)

    assert epg.get_channel_by_sid(2002).xmltv_id == 'f3932e75f691561adbe3b609369e487b'
    assert epg.get_channel_by_sid(2002).xmltv_display_name == 'BBC One Lon'
    assert epg.get_channel_by_sid(2002).xmltv_icon_url.human_repr() == 'http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png'

    # check that programme list is in order.
    start_times = [p.start for p in epg.get_channel_by_sid(2002).programmes]
    for index, time in enumerate(start_times):
        if index < len(start_times) - 1:
            assert start_times[index + 1] > time


def test_from_json():
    epg = EPG('localhost')
    json_file = Path(__file__).resolve().parent.joinpath('epg.json')

    with open(json_file, 'r') as jf:
        epg_json_data = jf.read()

    epg.from_json(epg_json_data)

    as_json = epg.as_json()

    assert epg.from_json(as_json) == epg.from_json(epg_json_data)

    assert epg.get_channel_by_sid(2002).c == "101"
    assert epg.get_channel_by_sid(2002).t == "BBC One Lon"
    assert epg.get_channel_by_sid(2002).name == "BBC One Lon"

    assert isinstance(epg.get_channel_by_sid(2002).programmes[1], Programme)
    assert epg.get_channel_by_sid(2002).programmes[1].title == "Rip Off Britain"

    assert epg.get_channel_by_sid(2002).programmes[1].start == parse("20181005091500 +0100")



import logging
import sys
from datetime import datetime
from functools import partial
from pathlib import Path

import pytest
import trio
from yarl import URL

from pyskyq import XMLTVListing

from .http_server import http_server
from .isloated_filesystem import isolated_filesystem

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.WARNING, stream=sys.stdout,
                    format=logformat)

LOGGER = logging.getLogger(__name__)

def test_xmltvlisting_init():
    # used to set up and tear down a temp dir for these tests.
    with isolated_filesystem():
        with pytest.raises(TypeError, match='path must be a string or Path object.'):
            l = XMLTVListing('http://blah.com/feed/6715', 6)

        Path.cwd().joinpath('somedir').mkdir()  # test the case where the directory is already there.
        m = XMLTVListing('http://blah.com/feed/6715', 'somedir')
        assert m._url == URL('http://blah.com/feed/6715')
        assert isinstance(m._path, Path)
        assert str(m._path) == 'somedir'
        assert "<XMLTVListing: url='http://blah.com/feed/6715', path='somedir'" in m.__repr__()
        assert m._path.is_dir()
        assert m.url == URL('http://blah.com/feed/6715')

        n = XMLTVListing(URL('http://blah.com/feed/6715'), '.str_path')
        assert n._url == URL('http://blah.com/feed/6715')
        assert isinstance(n._path, Path)
        assert str(n._path) == '.str_path'

        assert m == n
        assert m._filename == n._filename
        assert m.__hash__() == n.__hash__()


xmlfile_path = Path(__file__).resolve().parent.joinpath('parse_xmltv_data.xml')


async def test_download():
    with isolated_filesystem():
        with trio.move_on_after(5):
            async with trio.open_nursery() as server_nursery:
                LOGGER.debug('in nursery')
                async with await trio.open_file(xmlfile_path, 'r') as fd:
                    data = await fd.read()
                responses = [
                    {
                        'target': '/feed/6715',
                        'status_code': 200,
                        'content_type': "application/xml",
                        'body': bytearray(data.encode('utf-8')),
                        'headers': {
                            'Last-Modified': 'Mon, 08 Oct 2018 01:50:19 GMT',
                            'Connection': 'close',
                        },

                    },
                ]

                await server_nursery.start(trio.serve_tcp, partial(http_server, responses=responses), 8000)
                LOGGER.debug('started server.')


                l = XMLTVListing('http://localhost:8000/feed/6715')
                assert not l._downloaded
                await l.fetch()
                assert l._downloaded
                assert l.file_path.is_file()
                assert l.last_modified == datetime(2018, 10, 8, 1, 50, 19, 0)

                with open(xmlfile_path, 'rb') as src, open(l.file_path, 'rb') as dest:
                    assert src.read(-1) == dest.read(-1)


async def test_download_no_last_modified():
    with isolated_filesystem():
        with trio.move_on_after(5):
            async with trio.open_nursery() as server_nursery:
                LOGGER.debug('in nursery')
                async with await trio.open_file(xmlfile_path, 'r') as fd:
                    data = await fd.read()
                responses = [
                    {
                        'target': '/feed/6715',
                        'status_code': 200,
                        'content_type': "application/xml",
                        'body': bytearray(data.encode('utf-8')),
                        'headers': {
                            'Connection': 'close',
                        },

                    },
                ]

                await server_nursery.start(trio.serve_tcp, partial(http_server, responses=responses), 8000)
                LOGGER.debug('started server.')

                l = XMLTVListing('http://localhost:8000/feed/6715')
                assert not l._downloaded
                await l.fetch()
                assert l._downloaded
                assert l.file_path.is_file()
                assert l.last_modified is None

                with open(xmlfile_path, 'rb') as src, open(l.file_path, 'rb') as dest:
                    assert src.read(-1) == dest.read(-1)


xmlfile_path_gz = Path(__file__).resolve().parent.joinpath('parse_xmltv_data.xml.gz')


async def test_download_gzipped():
    with isolated_filesystem():
        with trio.move_on_after(5):
            async with trio.open_nursery() as server_nursery:
                LOGGER.debug('in nursery')
                async with await trio.open_file(xmlfile_path_gz, 'rb') as fd:
                    data = await fd.read()
                responses = [
                    {
                        'target': '/feed/6715',
                        'status_code': 200,
                        'content_type': "application/xml",
                        'body': data,
                        'headers': {
                            'Last-Modified': 'Mon, 08 Oct 2018 01:50:19 GMT',
                            'Connection': 'close',
                            'Content-Encoding': 'gzip',
                        },

                    },
                ]

                await server_nursery.start(trio.serve_tcp, partial(http_server, responses=responses), 8000)
                LOGGER.debug('started server.')

                l = XMLTVListing('http://localhost:8000/feed/6715')
                assert not l._downloaded
                await l.fetch()
                assert l._downloaded
                assert l.file_path.is_file()
                assert l.last_modified == datetime(2018, 10, 8, 1, 50, 19, 0)

                # note, we assert equivalence with the unzipped file to verify uncompression worked.
                with open(xmlfile_path, 'rb') as src, open(l.file_path, 'rb') as dest:
                    assert src.read(-1) == dest.read(-1)




#TODO -- add bad xml file test here.
def test_channel_parse():
    l = XMLTVListing('http://foo.com/feed/6715')

    with pytest.raises(OSError, match='File not downloaded, or download is currently in flight.'):
        for chan in l.parse():
            pass # should throw an error as file is not downloaded.

    l._full_path = Path(__file__).resolve().parent.joinpath('parse_xmltv_data.xml')
    l._downloaded = True

    for i,chan in enumerate(l.parse()):

        LOGGER.debug(chan.xmltv_id)
        LOGGER.debug(chan.xmltv_icon_url)
        LOGGER.debug(chan.xmltv_display_name)

        if i == 0:
            assert chan.xmltv_id == 'f3932e75f691561adbe3b609369e487b'
            assert chan.xmltv_display_name == 'BBC One Lon'
            assert chan.xmltv_icon_url.human_repr() == 'http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png'
        elif i == 1:

            assert chan.xmltv_id == 'a3c70f4c25110a9ca84f7c604023ee6c'
            assert chan.xmltv_display_name == 'Dave'
            assert chan.xmltv_icon_url.human_repr() == 'http://www.xmltv.co.uk/images/channels/a3c70f4c25110a9ca84f7c604023ee6c.png'
        else:
            break



"""This package aims to provide APIs for the SkyQ set-top box.

It aims to provide access to all functions that are exposed over the network by
a SkyQ box, and a few additional things besides.

Current SkyQ Boxes expose the following functions (that I'm aware of):

    - a uPNP interface for getting and setting certain settings.
    - a TCP socket interface that provides a remote control API.
    - a HTTP REST interface, for miscellaneous channel, status and EPG data.
    - a WebSocket JSON interface for subscribing to state changes, like Standby Mode.

Sky also provides some services to their SkyQ boxes over the internet, so
this API might tap into some of that stuff too. It also is capable of fetching
data from 3rd party XMLTV sources.

As of November 2018, this is still very much a work in progress.
"""

import logging

from .constants import REMOTECOMMANDS, CHANNELSOURCES, QUALITY
from ._version import __version__
from .remote import press_remote
from .epg import EPG
from .status import get_status
from .xmltvlisting import XMLTVListing
from .channel import (Channel, channel_from_skyq_service,
                      channel_from_xmltv_list, merge_channels,
                      channel_from_json)
from .programme import Programme, programme_from_xmltv_list

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""Misc util functions."""

import email.utils as eut
from datetime import datetime

from yarl import URL


def parse_http_date(httpdatetime: str) -> datetime:
    """Turn an HTTP-formated date-time string into a native :py:class:`~datetime.datetime` object.

    Args:
        httpdatetime (str): A string like ``Mon, 08 Oct 2018 01:42:20 GMT``

    Returns:
        datetime: A datetime object.

    """
    return datetime(*eut.parsedate(httpdatetime)[:6])  # type: ignore


#TODO: this is ugly, fix at some point.
def skyq_json_decoder_hook(obj):
    """Decode JSON into appropriate types used in this library."""
    if "xmltv_icon_url" in obj.keys():
        obj['xmltv_icon_url'] = URL(obj['xmltv_icon_url'])
    return obj



"""This module implements the EPG class."""

import json
import logging
from typing import Any

import asks
import trio

from .channel import (Channel, _ChannelJSONEncoder, channel_from_json,
                      channel_from_skyq_service, merge_channels)
from .constants import (REST_PORT, REST_SERVICE_DETAIL_URL_PREFIX,
                        REST_SERVICES_URL)
from .programme import Programme
from .xmltvlisting import XMLTVListing

LOGGER = logging.getLogger(__name__)

asks.init('trio')

class EPG:
    """The top-level class for all EPG data and functions.

    EPG implements access to any data and operations related to the SkyQ box's
    Electronic Programme Guide through a REST endpoint exposed on the box.
    Specifically, it fetches both summary and detail channel information from
    two different endpoints from the box and aggregates this data into a list of
    :class:`~.channel.Channel` objects.

    This channel data can then be *augmented* with data from an XMLTV feed site
    like http://www.xmltv.co.uk/ which provides an XML file which follows the ``xmltv``
    DTD.

    It also uses the ``xmltv`` file to load **listings** data to enable the
    querying of programmes that are scheduled on the channels.

    Args:
        host (str): String with resolvable hostname or IPv4 address to SkyQ box.
        rest_port (int): Defaults to the SkyQ REST port which is 9006.

    Attributes:
        host (str): Hostname or IPv4 address of SkyQ Box.
        rest_port (int): Port number to use to connect to the REST HTTP server.
            Defaults to the standard port used by SkyQ boxes which is 9006.


    """

    def __init__(self,
                 host: str,
                 *,
                 rest_port: int = REST_PORT,
                 ) -> None:
        """Initialise Sky EPG Object."""
        self.host: str = host
        self.rest_port: int = rest_port
        self._channels: list = []
        LOGGER.debug(f"Initialised EPG object using SkyQ box={self.host}")

    def __repr__(self):
        """Print a human-friendly representation of this object."""
        return f"<EPG: host={self.host}, rest_port={self.rest_port}, " + \
            f"len(_channels)={len(self._channels)}, len(_xmltv_urls)={len(self._channels)}>"

    def channels_loaded(self) -> bool:
        """Return whether channels have been loaded.

        Returns:
            bool: ``True`` if channels have been loaded.

        """
        return bool(self._channels)

    async def load_skyq_channel_data(self) -> None:
        """Load channel data into channel property.

        This method fetches the channel list from ``/as/services`` endpoint and
        loads the fetched channels into this ``EPG`` object. It then
        concurrently fetches each channel's detail info from
        ``/as/services/details/<sid>`` and adds that to the channels.

        Returns:
            None

        """
        async def _load_chan_detail(detail_url: str, channel: Channel) -> None:
            """Async function to fetch the details of each channel."""
            LOGGER.debug(f'Fetching channel details from {detail_url}')
            detail_payload_json = await sess.get(detail_url)
            LOGGER.debug(f'Got channel details from {detail_url}')
            detail_payload = detail_payload_json.json()
            LOGGER.debug(f'Parsed JSON from {detail_url}')
            channel.load_skyq_detail_data(detail_payload)
            LOGGER.debug(f'Merged {detail_url} into {channel}...')
            self._channels.append(channel)
            LOGGER.debug(f'Added {channel} to EPG')

        url = f'http://{self.host}:{self.rest_port}{REST_SERVICES_URL}'
        LOGGER.debug(f'Fetching channel list from {url}')
        sess = asks.Session(connections=50)
        LOGGER.debug(f'About to fetch channel list from sky box.')
        chan_payload_json = await sess.get(url)
        LOGGER.debug(f'Got channel list from sky box.')
        chan_payload = chan_payload_json.json()
        LOGGER.debug(f'Parsed channel list JSON payload.')

        with trio.move_on_after(90): # 90 sec timeout  #TODO parametrise this.
            async with trio.open_nursery() as nursery:
                LOGGER.debug('Started nursery to collect channel details.')
                for single_channel_payload in chan_payload['services']:
                    LOGGER.debug(f'Processing {single_channel_payload}')
                    channel = channel_from_skyq_service(single_channel_payload)
                    sid = channel.sid
                    detail_url = f'http://{self.host}:{self.rest_port}' + \
                        f'{REST_SERVICE_DETAIL_URL_PREFIX}{sid}'
                    nursery.start_soon(_load_chan_detail, detail_url, channel)


    def as_json(self) -> str:
        """Return the channel and programme data as JSON.

        Returns:
            str: A JSON representation of this EPG.

        """
        return json.dumps(self._channels, cls=_ChannelJSONEncoder, indent=4)

    def from_json(self, json_: str) -> None:
        """Load channel and programme data from JSON.

        Args:
            json_(str): A string of JSON text.

        """
        payload = json.loads(json_)
        new_chans: list = []

        for item in payload:
            if item['__type__'] == "__channel__":
                new_chans.append(channel_from_json(json.dumps(item)))
            # TODO: add jobs, programmes, etc
        self._channels = new_chans



    def get_channel_by_sid(self,
                           sid: Any
                           ) -> Channel:
        """Get channel data by SkyQ service id.

        This method returns a :class:`~.channel.Channel` object when
        passed in a channel ``sid``.

        Args:
            sid: The SkyQ service id of the channel
        Returns:
            :class:`~.channel.Channel`: The channel if found.

        Raises:
            ValueError: If the channel is not found or the channel
                list is empty.

        """
        sid = str(sid)
        if not self.channels_loaded():
            raise ValueError("No channels loaded.")

        for chan in self._channels:
            if chan.sid == sid:
                return chan
        raise ValueError(f"Sid:{sid} not found.")


    def apply_XMLTVListing(self, listing: XMLTVListing) -> None:
        """Merge listing data into the EPG channels data structure by channel name.

        This method takes an XMLTV listing object and, for every channel already loaded
        in the EPG it augments that channel's data with the data obtained from the XMLTV
        listing. This could provide additional descriptive channel data, the URL to the
        channel's logo and similar extra stuff.

        It then adds Programme objects to the Channel's Programme data structure.

        Args:
            listing(XMLTVListing): a :class:`~.xmltvlisting.XMLTVListing` object to
                merge to the EPG.

        Returns:
            None

        Raises:
            ValueError: Raised if the XMLTVListing file cannot be found.

        """
        if not listing.downloaded:
            raise ValueError("No XMLTVListing file found.")

        # There is no guarantee that the XMLTV file will serve elements in a particular order.
        # We want to load and create _all_ the channels before we load the programmes so we use
        # a temp var to hold the programme items which we then add to the Channels _afterwards_.

        prog_temp = []
        for xmltv_item in listing.parse():
            if isinstance(xmltv_item, Channel):
                sky_channel_names = [chan.name.lower() for chan in self._channels]
                if xmltv_item.xmltv_display_name.lower() in sky_channel_names:
                    idx = sky_channel_names.index(xmltv_item.xmltv_display_name.lower())
                    new_chan = merge_channels(self._channels[idx], xmltv_item)
                    LOGGER.debug(f'New channel:{new_chan}.')
                    self._channels.append(new_chan)
                    LOGGER.debug(f'Replaced {self._channels[idx]} with {new_chan} in EPG.')
                    del self._channels[idx]
                    continue
            if isinstance(xmltv_item, Programme):
                prog_temp.append(xmltv_item)
                continue

        # We rely on the Programme's hash being the same as any previous
        # instance of this programme that is being added to the programmes
        # SortedSet as the __hash__() function relies _only_ on the programm's
        # Title, StartTime and XMLChannelID data. In this way, if we apply a
        # newer, richer XMLTV listing to an existing one, by adding this item to
        # the SortedSet we will _replace_ any previously present Programme if
        # the hash of the newer one is the same as an existing one, even if the
        # data is newer/better. This allows us to keep applying XMLTV Listing
        # over and over without risk of duplication.
        for chan in self._channels:
            chan_progs = [prog for prog in prog_temp if prog.channel_xml_id == chan.xmltv_id]
            chan.programmes.update(chan_progs)



# This code stolen from https://h11.readthedocs.io/en/latest/examples.html
# and modified to suit my mocking needed.
# Credit to Nathaniel J. Smith for producting the original version.
# pylint: disable=no-member
import json
from itertools import count
from wsgiref.handlers import format_date_time

import trio
from trio_websocket import serve_websocket, ConnectionClosed
import h11
import logging

from functools import partial
import sys
MAX_RECV = 2 ** 16

TIMEOUT = 10

LOGGER = logging.getLogger(__name__)


class TrioHTTPWrapper:
    _next_id = count()

    def __init__(self, stream):
        self.stream = stream
        self.conn = h11.Connection(h11.SERVER)
        # Our Server: header
        self.ident = " ".join([
            "h11-example-trio-server/{}".format(h11.__version__),
            h11.PRODUCT_ID,
        ]).encode("ascii")
        # A unique id for this connection, to include in debugging output
        # (useful for understanding what's going on if there are multiple
        # simultaneous clients).
        self._obj_id = next(TrioHTTPWrapper._next_id)

    async def send(self, event):
        # The code below doesn't send ConnectionClosed, so we don't bother
        # handling it here either -- it would require that we do something
        # appropriate when 'data' is None.
        assert type(event) is not h11.ConnectionClosed
        data = self.conn.send(event)
        await self.stream.send_all(data)

    async def _read_from_peer(self):
        if self.conn.they_are_waiting_for_100_continue:
            LOGGER.debug("Sending 100 Continue")
            go_ahead = h11.InformationalResponse(
                status_code=100,
                headers=self.basic_headers())
            await self.send(go_ahead)
        try:
            data = await self.stream.receive_some(MAX_RECV)
        except ConnectionError:
            # They've stopped listening. Not much we can do about it here.
            data = b""
        self.conn.receive_data(data)

    async def next_event(self):
        while True:
            event = self.conn.next_event()
            if event is h11.NEED_DATA:
                await self._read_from_peer()
                continue
            return event

    async def shutdown_and_clean_up(self):
        # When this method is called, it's because we definitely want to kill
        # this connection, either as a clean shutdown or because of some kind
        # of error or loss-of-sync bug, and we no longer care if that violates
        # the protocol or not. So we ignore the state of self.conn, and just
        # go ahead and do the shutdown on the socket directly. (If you're
        # implementing a client you might prefer to send ConnectionClosed()
        # and let it raise an exception if that violates the protocol.)
        #
        try:
            await self.stream.send_eof()
        except trio.BrokenResourceError:
            # They're already gone, nothing to do
            return
        # Wait and read for a bit to give them a chance to see that we closed
        # things, but eventually give up and just close the socket.
        # XX FIXME: possibly we should set SO_LINGER to 0 here, so
        # that in the case where the client has ignored our shutdown and
        # declined to initiate the close themselves, we do a violent shutdown
        # (RST) and avoid the TIME_WAIT?
        # it looks like nginx never does this for keepalive timeouts, and only
        # does it for regular timeouts (slow clients I guess?) if explicitly
        # enabled ("Default: reset_timedout_connection off")
        with trio.move_on_after(TIMEOUT):
            try:
                while True:
                    # Attempt to read until EOF
                    got = await self.stream.receive_some(MAX_RECV)
                    if not got:
                        break
            finally:
                await self.stream.aclose()

    def basic_headers(self):
        # HTTP requires these headers in all responses (client would do
        # something different here)
        return [
            ("Date", format_date_time(None).encode("ascii")),
            ("Server", self.ident),
        ]


async def http_server(stream, responses):
    wrapper = TrioHTTPWrapper(stream)
    while True:
        assert wrapper.conn.states == {
            h11.CLIENT: h11.IDLE, h11.SERVER: h11.IDLE}

        try:
            with trio.fail_after(10):
                LOGGER.debug("Server main loop waiting for request")
                event = await wrapper.next_event()
                LOGGER.debug(f"Server main loop got event: {event}")
                if isinstance(event, h11.Request):
                    response = [response for response in responses if response['target'].encode('ascii') == event.target][0]
                    await send_mock_response(wrapper, event, response)
        except Exception as exc:
            LOGGER.debug(f"Error during response handler: {exc}")
            await maybe_send_error_response(wrapper, exc)

        if wrapper.conn.our_state is h11.MUST_CLOSE:
            LOGGER.debug("connection is not reusable, so shutting down")
            await wrapper.shutdown_and_clean_up()
            return
        else:
            try:
                LOGGER.debug("trying to re-use connection")
                wrapper.conn.start_next_cycle()
            except h11.ProtocolError:
                states = wrapper.conn.states
                LOGGER.debug(f"unexpected state: {states}, -- bailing out")
                await maybe_send_error_response(
                    wrapper,
                    RuntimeError("unexpected state {}".format(states)))
                await wrapper.shutdown_and_clean_up()
                return


################################################################
# Actual response handlers
################################################################

# Helper function


async def send_simple_response(wrapper, status_code, content_type, body, passed_headers):
    LOGGER.debug(f"Sending {status_code} response with {len(body)} bytes")
    headers = wrapper.basic_headers()
    headers.append(("Content-Type", content_type))
    headers.append(("Content-Length", str(len(body))))
    if passed_headers:
        for k, v in passed_headers.items():
            headers.append((k, v))
    LOGGER.debug(f'Final headers = {headers}')

    res = h11.Response(status_code=status_code, headers=headers)
    await wrapper.send(res)
    await wrapper.send(h11.Data(data=body))
    await wrapper.send(h11.EndOfMessage())


async def maybe_send_error_response(wrapper, exc):
    # If we can't send an error, oh well, nothing to be done
    LOGGER.debug("trying to send error response...")
    if wrapper.conn.our_state not in {h11.IDLE, h11.SEND_RESPONSE}:
        LOGGER.debug(f"...but I can't, because our state is: {wrapper.conn.our_state}")
        return
    try:
        if isinstance(exc, h11.RemoteProtocolError):
            status_code = exc.error_status_hint
        elif isinstance(exc, trio.TooSlowError):
            status_code = 408  # Request Timeout
        else:
            status_code = 500
        body = str(exc).encode("utf-8")
        await send_simple_response(wrapper,
                                   status_code,
                                   "text/plain; charset=utf-8",
                                   body,
                                   passed_headers=None) # no special headers needed.
    except Exception as exc:
        LOGGER.debug(f"error while sending error response: {exc}")


async def send_mock_response(wrapper, request, response):
    LOGGER.debug("Preparing mock response")
    if request.method not in {b"GET", b"POST"}:
        # Laziness: we should send a proper 405 Method Not Allowed with the
        # appropriate Accept: header, but we don't.
        raise RuntimeError("unsupported method")
    await send_simple_response(wrapper,
                               response['status_code'],
                               response['content_type'],
                               response['body'],
                               response['headers']
                               )


async def websocket_request_handler(request, *, responses):
    ws = await request.accept()
    for res in responses:
        try:
            await ws.send_message(res['body'])
            await trio.sleep(0.2)
        except ConnectionClosed:
            continue


async def websocket_server(ip_addr, port, responses):
    await serve_websocket(partial(websocket_request_handler, responses=responses), ip_addr, port, ssl_context=None)



"""This module implements the ``press_remote()`` function.

Remote is a lower level interface to the SkyQ box that seeks to emulate most of the the
button presses available on the SkyQ remote. There are some buttons which are not (currently)
supported, however.

Todos:
    * Research some of the missing button codes to see if it's possible to also include them.
"""

import logging
import socket
import math

from .constants import REMOTE_PORT

LOGGER = logging.getLogger(__name__)

def press_remote(host: str, code: int, *, port: int = REMOTE_PORT) -> None:
    """Send a command to the Sky Q box using the Remote interface.

    This function sends a button press to the SkyQ box via a TCP socket connection. The
    encoding and handshake required was extracted by looking at the following sources:

    - https://github.com/dalhundal/sky-remote
    - https://gladdy.uk/blog/2017/03/13/skyq-upnp-rest-and-websocket-api-interfaces/

    Args:
        host (str): String with resolvable hostname or IPv4 address to SkyQ box.
        code (int): A code passed which represents the button to press. See
            :py:mod:`.constants` for a human-friendly list of the currently
            supported buttons and their associated codes.
        port (int, optional): Port number to use to connect to the Remote TCP socket.
            Defaults to the standard port used by SkyQ boxes which is 49160.

    Returns:
        None

    """
    LOGGER.debug(f'Sending command code={code}')
    command_bytes = bytearray([4, 1, 0, 0, 0, 0, math.floor(224 + (code/16)), code % 16])
    LOGGER.debug(f'command_bytes={command_bytes}')

    # TODO try/except socket.gaierror
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect((host, port))

    # Wait until ready for data
    length = 12
    while True:
        data = client.recv(24)
        LOGGER.debug(f'Received data={data}')
        if len(data) < 24: # box sends 24 x \x00 to signal end of hand-shake.
            client.send(data[0:length])
            LOGGER.debug(f'Sent data={data[0:length]}')
            length = 1
        else:
            break

    client.send(command_bytes)
    LOGGER.debug(f'Send command part 1 data={command_bytes}')
    command_bytes[1] = 0
    client.send(command_bytes)
    LOGGER.debug(f'Send command part 2 data={command_bytes}')
    client.close()



#!/usr/bin/env python
"""Cli wrapper for the Remote API.

Example:
    You can press a button on the SkyQ remote like this:

        ``$ python cli_remote.py pause``

        ``$ python cli_remote.py play``
"""

import argparse
import sys
from typing import List

from pyskyq import press_remote
from pyskyq import REMOTECOMMANDS as RCMD

def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse command line parameters."""
    parser = argparse.ArgumentParser(
        description="Presses a button on the SkyQ remote")
    parser.add_argument(
        dest="cmd",
        help="command to send",
        type=str,
        metavar="CMD")
    return parser.parse_args(args)


def main(args: List[str]):
    """Provide main entry point."""
    pargs = parse_args(args)
    # replace 'skyq' with hostname / IP of your Sky box
    press_remote('skyq', getattr(RCMD, pargs.cmd))


if __name__ == "__main__":
    main(sys.argv[1:])



"""This module implements the XMLTVListing listing class."""
#pylint: disable=line-too-long
import hashlib
import logging
import shutil
import zlib
from collections.abc import Hashable
from datetime import datetime
from functools import partial
from os import PathLike
from pathlib import Path
from typing import IO, Any, Iterator, Optional, Tuple, Union
from xml.etree.ElementTree import Element, iterparse

import asks
import trio
from yarl import URL

from .channel import channel_from_xmltv_list, Channel
from .programme import programme_from_xmltv_list, Programme
from .utils import parse_http_date

LOGGER = logging.getLogger(__name__)
asks.init('trio')


class XMLTVListing(Hashable):  # pylint: disable=too-many-instance-attributes
    """Hold the data and functions required to obtain an XMLTVListing.

    Currently it only supports the XML TV Listings format which holds
    Channel and Programme scheduling information. See the dtd_ for details.

    .. _dtd: https://github.com/AlekSi/xmltv/blob/master/xmltv.dtd

    Note:

        The :class:`~.channel.Channel` class provides the **primary**
        interface to channel data through the :class:`~.epg.EPG` class.

        This class provides the means to download and parse XML data to do with
        channels, but more importantly, programming schedules. While it can be used
        stand-alone, it is most effective when injected into the
        :class:`~.epg.EPG` object using
        :meth:`.epg.EPG.apply_XMLTVListing()`.

    """

    def __init__(self,
                 url: URL,
                 path: PathLike = Path('.epg_data'),
                 ) -> None:
        """Instantiate the object with the URL to fetch."""
        self._url: URL
        self._path: Path

        if not isinstance(url, URL):
            self._url = URL(url)
        else:
            self._url = url

        # Validate path
        if not isinstance(path, Path) and not isinstance(path, str):
            raise TypeError('path must be a string or Path object.')

        # coerce it to a Path if it's a string
        if isinstance(path, str):
            self._path = Path(path)
        else:
            self._path = path

        if not self._path.is_dir():
            self._path.mkdir()

        str_binary = str(self._url.human_repr()).encode('utf8')
        hashobj = hashlib.sha256(str_binary)
        self._hash = int.from_bytes(hashobj.digest(), 'big')
        self._filename = f'{hashobj.hexdigest()}.xml'
        self._full_path = self._path.joinpath(self._filename)
        self._last_modified: Optional[datetime] = None
        self._downloaded: bool = False
        self._downloading: bool = False

        LOGGER.debug(f'XMLTVListing initialised: {self}')

    def __hash__(self) -> int:
        """Define hash function for a Hashable object."""
        return self._hash

    def __eq__(self, other) -> bool:
        """Define equality test for a Hashable object."""
        # relying on lazy boolean evaluation here.
        return isinstance(other, XMLTVListing) and hash(other._url) == hash(self._url)  # pylint: disable=protected-access

    def __repr__(self):
        """Print a human-friendly representation of this object."""
        return f"<XMLTVListing: url='{self._url}', path='{self._path}', filename='{self._filename}'>"

    @property
    def last_modified(self):
        """Return the last modified date from the HTTP header of the last download.

        Returns the date and time when the data was last modified. Taken directly
        from the ``HTTP-Header``.

        If the header was not provided, it returns ``None``.

        Returns:
            datetime: A :py:class:`datetime.datetime` object or ``None``.

        """
        return self._last_modified

    @property
    def url(self) -> URL:
        """Return the url of this XMLTVListing.

        Returns:
            :py:class:`yarl.URL`

        """
        return self._url

    @property
    def downloaded(self) -> bool:
        """Return the status of XMLTV file download.

        Returns:
            bool: ``True`` if the file was downloaded successfully

        """
        return self._downloaded

    @property
    def downloading(self) -> bool:
        """Return the status of XMLTV file download.

        Returns:
            bool: ``True`` if the file is currently being downlaoded

        """
        return self._downloading

    @property
    def file_path(self) -> Path:
        """Return the full file path of this listing's XML file.

        Returns:
            :py:class:`pathlib.Path`: A `Path` to the location of the
                XML file (whether it has yet been :meth:`fetch`'ed or not).

        """
        return self._full_path

    # TODO -- add error handling for
    # -- HTTP headers missing
    # -- timeouts
    # -- badURL, etc, etc.
    # TODO -- add retry support

    async def fetch(self) -> None:
        """Fetch the XMLTVListing file.

        This async method will download the (XML) file specified by the URL passed
        at instantiation.

        If the server does not support streaming downloads the method will will
        fall back to a more memory-intensive, vanilla HTTP download.

        Args:
            timeout (int): timeout in seconds for the HTTP session. Defaults to
                60 seconds.
            range_size (int): the size, in bytes, of each chunk. Defaults to
                256k (256*1024 bytes).

        Returns:
            None

        """
        LOGGER.debug(f'Fetch({self}) call started.')
        self._downloading = True
        newfile = self._full_path.with_suffix('.tmp')


        # We support gzip encoding as www.xmltv.co.uk sends stuff gzipped.
        # the zlib decompressiion object supports streaming decompression which
        # why we use it.

        # We should probably support more encoding algorithms at some point.

        decompression_obj = zlib.decompressobj(wbits=16 + zlib.MAX_WBITS)
        compressed_payload = True  # assume payload is compressed first.

        async def chunk_processor(decomp_obj, bytechunk):
            nonlocal compressed_payload
            async with await trio.open_file(newfile, 'ab') as output_file:
                if compressed_payload:
                    try:
                        LOGGER.debug(f'Got zipped chunk. size = {len(bytechunk)}. Unzipping...')
                        await output_file.write(decomp_obj.decompress(bytechunk))
                        LOGGER.debug(f'Wrote unzipped chunk. size = {len(bytechunk)}')
                    except zlib.error:
                        await output_file.write(bytechunk)
                        LOGGER.debug(f'Wrote chunk. size = {len(bytechunk)}')
                        compressed_payload = False  # it wasn't compressed, so fall back
                else:
                    await output_file.write(bytechunk)
                    LOGGER.debug(f'Wrote chunk. size = {len(bytechunk)}')


        resp = await asks.get(str(self._url),
                              headers={'Accept-Encoding': 'gzip'},
                              # callback takes bytes only
                              callback=partial(chunk_processor, decompression_obj)
                              )

        LOGGER.debug(f'Got headers = {resp.headers}')

        # h11 makes header keys lowercase so we can reply on this
        if resp.headers.get("last-modified"):
            self._last_modified = parse_http_date(resp.headers.get("last-modified"))
            LOGGER.debug(f'Content last modified on: {resp.headers.get("last-modified")}.')

        shutil.move(newfile, self._full_path)
        self._downloading = False
        self._downloaded = True
        LOGGER.debug(f'Fetch finished on {self}')


    def parse(self) -> Iterator[Union[Channel, Programme]]:
        """Parse the XMLTVListing XML file and create an iterator over the data in it.

        Yields:
            (Any): Either a  :class:`~pyskyq.channel.Channel` or a
                :class:`~pyskyq.programme.Programme` object is yielded.

        """
        if not self.downloaded and not self.downloading:
            raise OSError('File not downloaded, or download is currently in flight.')
        else:
            LOGGER.debug(f'in parse_channels. file = {self.file_path}')
            for xml_type, xml_item in _xml_parse_and_remove(self._full_path): # type: ignore
                if xml_type == 'channel':
                    LOGGER.debug('yielding channel...')
                    yield channel_from_xmltv_list(xml_item)
                if xml_type == 'programme':
                    LOGGER.debug('yielding programme...')
                    yield programme_from_xmltv_list(xml_item)

# TODO: add error checking for XML errors
# TODO: see if this can be made async.
def _xml_parse_and_remove(filename: Union[str, bytes, int, IO[Any]],
                          ) -> Iterator[Tuple[str, Element]]:
    """Incrementally load and parse an XML file.

    Stolen from Python Cookbook 3rd edition, section 6.4 with credit to the book's authors.

    Args:
        filename(Union[str, bytes, int, IO[Any]]): a file-like object.
        path(str): The XML element to parse

    Yields:
        Element: A Parsed XML element.

    """
    doc = iterparse(filename, ('start', 'end'))
    # skip the root element
    next(doc)  # pylint: disable=stop-iteration-return
    tag_stack = []
    elem_stack = []
    for event, elem in doc:
        if event == 'start':
            LOGGER.debug(f'Parsing XML element: {elem.tag}')
            tag_stack.append(elem.tag)
            elem_stack.append(elem)
        else:  # event == 'end'
            if tag_stack == ['channel']:
                yield 'channel', elem
                elem.clear()
            elif tag_stack == ['programme']:
                yield 'programme', elem
            try:
                tag_stack.pop()
                elem_stack.pop()
            except IndexError:
                pass



# pylint: skip-file

import asyncio
import socket
import sys

import aiohttp
import pytest
from asynctest import CoroutineMock, MagicMock

from pyskyq import REMOTECOMMANDS as RCMD

from .asynccontextmanagermock import AsyncContextManagerMock
from .mock_constants import REMOTE_TCP_MOCK


# Disabled for now -- to be refactored in https://gitlab.com/bradwood/pyskyq/issues/21

# def test_cli(mocker):

#     m = mocker.patch('socket.socket')
#     m.return_value.recv.side_effect = REMOTE_TCP_MOCK

#     a = mocker.patch('aiohttp.ClientSession.get', new_callable=AsyncContextManagerMock)
#     a.return_value.__aenter__.return_value.json = CoroutineMock(side_effect=SERVICE_MOCK)

#     mocker.patch.object(sys, 'argv', ['pyskyq', 'play'])

#     run()

#     m.assert_called_with(socket.AF_INET, socket.SOCK_STREAM)

#     assert m.return_value.recv.call_count == 4
#     m.return_value.send.assert_any_call(b'SKY 000.001\n')
#     m.return_value.send.assert_any_call(b'\x01')
#     m.return_value.send.assert_any_call(b'\x00')
#     m.return_value.send.assert_any_call(bytearray(b'\x04\x00\x00\x00\x00\x00\xe4\x00'))

#     # TODO test with --verbose



"""Monkey patch H11 to ensure that it sends headers the way the SkyQ box wants them to be."""

def write_headers_titlecase(headers, write):
    """Monkey-patch h11 header handling.

    See https://github.com/python-hyper/h11/issues/31 for the details. Essentially SkyQ box is
    not RFC compliant as regards case-insensitive headers, so this patch forces the headers to
    comply with SkyQ's braindead implementation.
    """
    # RFC says Host: header SHOULD be written out first.
    for name, value in headers:
        if name == b"host":
            write(b"%s: %s\r\n" % (name.title(), value))
    for name, value in headers:
        if name == b"upgrade":
            # SkyQ doesn't like 'WebSocket', it wants 'websocket'
            write(b"%s: %s\r\n" % (name.title(), value.lower()))
        elif name != b"host":
            write(b"%s: %s\r\n" % (name.title(), value))
    write(b"\r\n")



#!/usr/bin/env python
"""Show the use of the EPG, Channel, XMLTVListing and Programme objects."""
import logging
import argparse
import sys
from typing import List

import trio

from pyskyq import EPG, XMLTVListing

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.DEBUG, stream=sys.stdout,
                    format=logformat)



def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse command line parameters."""
    parser = argparse.ArgumentParser(
        description="Query Channel Description from SkyQ EPG")
    parser.add_argument(
        dest="sid",
        help="channel sid, try 2002 as an example",
        type=str,
        metavar="SID")
    return parser.parse_args(args)


async def main(args: List[str]):
    """Run main routine, allowing arguments to be passed."""
    pargs = parse_args(args)
    epg = EPG('10.0.1.6')  # replace with hostname / IP of your Sky box
    await epg.load_skyq_channel_data()  # load channel listing from Box.
    all_72_hour = XMLTVListing('http://www.xmltv.co.uk/feed/6715')
    #limited_7_day = XMLTVListing('http://www.xmltv.co.uk/feed/6784')

    async with trio.open_nursery() as nursery:
        # fetch 2 separate XMLTV listing concurrently.
        nursery.start_soon(all_72_hour.fetch)

    assert all_72_hour.downloaded
    print("Downloaded XMLTV Listing: {all_72_hour}")

    # assert limited_7_day.downloaded
    # print("Downloaded XMLTV Listing: {limited_7_day}")

    epg.apply_XMLTVListing(all_72_hour)

    #epg.apply_XMLTVListing(limited_7_day)

    print('Channel Description from the SkyQ Box:')
    print(epg.get_channel_by_sid(pargs.sid).desc)
    print('Channel XMLTV ID from the XMLTV Feed:')
    print(epg.get_channel_by_sid(pargs.sid).xmltv_id)
    print('Channel Logo URL from the XMLTV Feed:')
    print(epg.get_channel_by_sid(pargs.sid).xmltv_icon_url)

    print('\n\n')
    print('---- EPG Listing ----')
    for prog in epg.get_channel_by_sid(pargs.sid).programmes:
        print(f'Name: {prog.title}', end='')
        print(f' Start: {prog.start}')
        print(f'{prog.desc}')




if __name__ == "__main__":
    trio.run(main, sys.argv[1:])



#!/usr/bin/env python
"""Example use of the Status API."""

import logging
import sys
import trio

from pyskyq import get_status

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.WARNING, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

logging.getLogger().setLevel(logging.DEBUG)  # status changes are logged.

async def report_box_online():
    """Report whether the SkyQ is online or not."""
    # pylint: disable=not-async-context-manager
    async with get_status('skyq') as stat:
        while True:
            if stat.online:
                print('The SkyQ Box is Online ')
            else:
                print('The SkyQ Box is Offline')
            await trio.sleep(1)
try:
    print("Type Ctrl-C to exit.")
    trio.run(report_box_online)
except KeyboardInterrupt:
    raise SystemExit(0)



"""This module implements the Status class.

It schedules subscriptions/polling tasks that monitor various events emitted
from the SkyQ box. Each of these will await an event (e.g. on a websocket or
uPNP subscription) and then update object attributes when neccessary.
"""
import json
import logging
from contextlib import asynccontextmanager

import h11  # this import needed for the monkey-patch
import trio
from trio_websocket import open_websocket_url, WebSocketConnection

from dataclasses import dataclass, field

from ._h11_header_monkey_patch import write_headers_titlecase
from .constants import REST_PORT, REST_STATUS_URL

LOGGER = logging.getLogger(__name__)

# Monkey patch h11 HTTP header writing to override case. SkyQ needs this to be happy.
h11._writers.write_headers = write_headers_titlecase # pylint: disable=protected-access

@dataclass
class Status:
    """This dataclass provides real-time access to the status of the SkyQ box.

    Example:

        It should be access via an asynchronous context manager like so::

            async def report_box_online():
                async with get_status('skyq') as stat:
                    while True:
                        if stat.standby:
                            print('The SkyQ Box is in Standby Mode')
                        else:
                            print('The SkyQ Box is in Online Mode')
                        await trio.sleep(1)

            try:
                print("Type Ctrl-C to exit.")
                trio.run(report_box_online)
            except KeyboardInterrupt:
                raise SystemExit(0)

    Attributes:
        online (bool): Is the box in online?

    Todo:
        Add more attributes once the JSON has been figured out.

    """

    online: bool = field(init=False, default=False)


@asynccontextmanager
async def get_status(host: str,
                     *,
                     port: int = REST_PORT,
                     ws_url_path: str = REST_STATUS_URL,
                     ):
    """Yield a Status object through an asynchronous context manager.

    This async function yields a :class:`Status` object asynchronously to enable
    the calling code to interact with is while it is updated via a websocket connection
    to the SkyQ box.

    Args:
        host (str): Hostname/IP address of the SkyQ Box
        port (int): Port number to connect to. Defaults to
            :data:`~pyskyq.constants.REST_PORT`.
        w_url_path (str): URL path to the websocket. Defaults to
            :data:`~pyskyq.constants.REST_STATUS_URL`

    Yields:
        status (Status): A :class:`Status` object.

    """
    ws_url: str = f'ws://{host}:{port}{ws_url_path}'
    LOGGER.debug(f'About to connect to: {ws_url}')
    status = Status()

    # no try/except as we want the exception thown up the stack if the socket's closed.
    async with open_websocket_url(ws_url) as conn:
        LOGGER.debug(f'Created connection to socket: {ws_url}')
        async with trio.open_nursery() as ws_nursery:
            LOGGER.debug('In ws_nursery yielding loop...')
            LOGGER.debug('Backgrounding connection_handler...')
            ws_nursery.start_soon(_handle_connection, status, conn)
            LOGGER.debug(f'Handler backgrounded. Yielding...')
            yield status


async def _handle_connection(s: Status, conn: WebSocketConnection):
    LOGGER.debug(f'Entered connection handler...{conn}')
    async with trio.open_nursery() as handler_nursery:
        LOGGER.debug('Created handler_nursery...')
        LOGGER.debug('Backgrounded _process_messages...')
        handler_nursery.start_soon(_process_messages, s, conn)

# TODO parse the entire payload and load it into Status.
async def _process_messages(s: Status, conn: WebSocketConnection):
    LOGGER.debug('Inside _process_messages...')
    while True:
        LOGGER.debug(f'conn: {conn}')
        data_json = await conn.get_message()
        LOGGER.debug('Got data from websocket...')
        data = json.loads(data_json)
        LOGGER.debug(f'Parsed JSON data from websocket.')
        s.online = data['hdmi']['state'] == 'available'



from asynctest import MagicMock

class AsyncContextManagerMock(MagicMock):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for key in ('aenter_return', 'aexit_return'):
            setattr(self, key, kwargs[key] if key in kwargs else MagicMock())

    async def __aenter__(self):
        return self.aenter_return

    async def __aexit__(self, *args):
        pass
        # return self.aexit_return

    # # Grabbed this hack from here
    # # https://github.com/aaugustin/websockets/issues/359
    def __await__(self):
        print('__await__', self)
        return self
        # return self().__await__()
        # the mere presence of this unreachable code makes this look like a generator! WTF?!
        yield None  # pylint: disable=unreachable



#!/usr/bin/env python
# pylint: skip-file

import sys
from setuptools import setup


def setup_package():
    needs_sphinx = {'build_sphinx', 'upload_docs'}.intersection(sys.argv)
    sphinx = ['sphinx'] if needs_sphinx else []
    setup(setup_requires=['pyscaffold>=3.0a0,<3.1a0'] + sphinx,
          use_pyscaffold=True)


if __name__ == "__main__":
    setup_package()



# pylint: skip-file

import socket

import pytest

from pyskyq import REMOTECOMMANDS as RCMD
from pyskyq import press_remote

from .mock_constants import REMOTE_TCP_MOCK


def test_sky_remote_send_command(mocker):

    m = mocker.patch('socket.socket')
    m.return_value.recv.side_effect = REMOTE_TCP_MOCK

    press_remote('some_host', RCMD.play)

    m.assert_called_with(socket.AF_INET, socket.SOCK_STREAM)

    assert m.return_value.recv.call_count == 4
    m.return_value.send.assert_any_call(b'SKY 000.001\n')
    m.return_value.send.assert_any_call(b'\x01')
    m.return_value.send.assert_any_call(b'\x00')
    m.return_value.send.assert_any_call(bytearray(b'\x04\x00\x00\x00\x00\x00\xe4\x00'))



"""Constants for button codes, ports, and so on."""

from enum import IntEnum, Enum, IntFlag

REMOTE_LEGACY_PORT: int = 5900
"""int: Legacy port number for older Sky Boxes."""

REMOTE_PORT: int = 49160
"""int: Current port number for newer SkyQ Boxes."""

class QUALITY(Enum):
    """Enumeration of Video Quality constants."""

    UHD = 'uhd' # ultra-high def
    HD = 'hd'  # high def
    SD = 'sd'  # standard def
    AU = 'au'  # audio



class CHANNELSOURCES(IntFlag):
    """Enumeration of Channel Sources.

    These are ORed together when sources are overlaid.
    """

    no_source = 0
    skyq_service_summary = 1  # data got from skyq:9006/as/service
    skyq_service_detail = 2   # data got from skyq:9006/as/service/detail/<sid>
    xml_tv = 4  # data got from XMLTV servce
    all_source = skyq_service_summary | skyq_service_detail | xml_tv


class REMOTECOMMANDS(IntEnum):
    """Enumeration of Remote codes."""

    power = 0
    select = 1
    backup = 2
    dismiss = 2
    channelup = 6
    channeldown = 7
    interactive = 8
    sidebar = 8
    help = 9
    services = 10
    search = 10
    tvguide = 11
    home = 11
    i = 14
    text = 15
    up = 16
    down = 17
    left = 18
    right = 19
    red = 32
    green = 33
    yellow = 34
    blue = 35
    zero = 48
    one = 49
    two = 50
    three = 51
    four = 52
    five = 53
    six = 54
    seven = 55
    eight = 56
    nine = 57
    play = 64
    pause = 65
    stop = 66
    record = 67
    fastforward = 69
    rewind = 71
    boxoffice = 240
    sky = 241

REST_PORT: int = 9006
"""int: Port number for the REST API."""

REST_SERVICES_URL: str = '/as/services'
"""str: REST endpoint for list of services (channels)."""

REST_STATUS_URL: str = '/as/system/status'
"""str: REST web-socket endpoint for box status/"""

REST_SERVICE_DETAIL_URL_PREFIX: str = '/as/services/details/'
"""str: REST endpoint for service (channel) details."""

CHANNEL_FIELD_MAP = {
    "number": "c", # "101",
    "quality": "sf", # "sd",
    "id": "sid", # "2002",
    "name": "t",  # "BBC One Lon",
    "desc": "upgradeMessage",
}
"""Dict(str,str): Maps human-friendly names to Sky's terse field names (where understood)."""



"""This module implements the Channel class and associated factory functions."""
import json
import logging
from collections.abc import Hashable
from typing import Any, Dict
from xml.etree.ElementTree import Element

from sortedcontainers.sortedset import SortedSet
from yarl import URL

from .constants import CHANNEL_FIELD_MAP
from .constants import CHANNELSOURCES as CSRC
from .programme import Programme
from .utils import skyq_json_decoder_hook

LOGGER = logging.getLogger(__name__)


class _ChannelJSONEncoder(json.JSONEncoder):
    """Encode Channel objects in JSON."""

    def default(self, obj):  # pylint: disable=arguments-differ,method-hidden,inconsistent-return-statements
        if isinstance(obj, Channel):
            type_ = '__channel__'
            chan_dict = obj._chan_dict # pylint: disable=protected-access
            sources = obj._sources  # pylint: disable=protected-access
            programmes = obj.programmes  # pylint: disable=protected-access
            return {'__type__': type_,
                    'attributes': chan_dict,
                    'sources': sources,
                    'programmes': programmes,
                    }

        if isinstance(obj, SortedSet):
            return list(obj)

        if isinstance(obj, Programme):
            return {'__type__': '__programme__',
                    # remove attributes that are not needed to re-create the object.
                    'attributes': {k: v for k, v in vars(obj).items()
                                   if k not in ['start', 'stop', 'channel_xml_id']}
                    }

        if isinstance(obj, URL):
            return obj.human_repr()

        json.JSONEncoder.default(self, obj)  # pragma: no cover


class Channel(Hashable):
    """This class holds channel data and methods for manipulating the channel.

    Specifically, this includes information about a particular channel and a SortedSet
    of programme data.

    Channel data can come form a variety of sources, including the SkyQ box
    itself or from a remote XML TV feed. Programme data relies upon the XMLTV
    Source data feed as I haven't (yet) figured out how to source Programme data
    from Sky/SkyQ directly yet.

    Warning:
        This class cannot be instantiated directly, but only via one of the factory methods.

    Note:
        Attributes are dynamically set based on the data payload provided, so if
        the SkyQ box or XML TV feed is upgraded, new attributes *should*
        magically appear. Note that this also implies that not every channel has
        every field, e.g., the ``timeshifted`` and ``adult`` properties are only
        present if they are  ``True``. This Sky weirdness is managed by simply
        returing ``None`` if the attribute is not present, rather than raising a
        ``KeyError``.

    Note:
        Data from XML TV sources presents the following channel attributes.

    Attributes:
        xmltv_id (str): A hex string uniquely identifying this channel in the XML TV listing.
            It is needed to look up programme's on the channel.
        xmltv_icon_url (:py:class:`yarl.URL`): An internet URL to a graphic logo file.
        xmltv_display_name (str): The human-friendly name of the channel. Should be the same
            as ``name``.
        programmes (OrderedSet[:class:`~pyskyq.programme.Programme`]): Holds a set of
            :class:`~pyskyq.programme.Programme` objects ordered by start time.

    Note:
        The following attributes come from the SkyQ Box's various endpoints.

    Attributes:
        number (str): Human-friendly version of :attr:`c`.
        quality (str): Human-friendly version of :attr:`sf`.
        id (str): Human-friendly version of :attr:`sid`.
        name (str):Human-friendly version of :attr:`t`.
        desc (str): Human-friendly version of :attr:`upgradeMessage`.
        c (str): Channel number.
        dvbtriplet (str):  DVB Triplet (I have no idea what this is)
        schedule (bool): Is the channel "scheduled" or not?
        servicetype (str): Where is the channel coming from (e.g., ``DSAT``)
        sf (str): Quality of the channel (e.g., ``hd``, ``sd``, or ``au``)
        sg (int): No ideas what this is.
        sid (str): Channel id (aka ``sid``) in `str` form - the **primary key** for the channel.
        sk (int): Channel id (aka ``sid``) in `int` form.
        t (str): Channel name, eg, ``BBC One Lon``
        xsg (int): No idea what this is.
        timeshifted (bool): Is this a ``+1``-style channel?
        adult (bool): Is this an adult channel?
        upgradeMessage (str): A short description of the channel.
        isbroadcasting (boo): Is the channel broadcasting currently or not?

    """

    def __init__(self) -> None:
        """Raise exception to prevent this from being initatiated normally."""
        raise NotImplementedError('This class cannot be instantiated directly. ' +
                                  'Use a factory function to create an instance.')

    def __new__(cls, *args, **kwargs):
        """Initialise Channel Object."""
        blank_chan_dict = {
            'sid': None,
            'c': None,
            't': None,
            'xmltv_id': None,
        }
        # pylint: disable=attribute-defined-outside-init
        new_obj = super(Channel, cls).__new__(cls, *args, **kwargs)
        new_obj._chan_dict: Dict[str, Any] = blank_chan_dict
        new_obj._sources: CSRC = CSRC.no_source
        new_obj.programmes: SortedSet = SortedSet()
        return new_obj

    def __getattr__(self, name: str) -> Any:
        """Handle attribute reads.

        Return ``None`` rather than a KeyError if the attriute not found.
        """
        if name in self._chan_dict.keys():
            return self._chan_dict[name]
        if name in CHANNEL_FIELD_MAP.keys():
            return self._chan_dict.get(CHANNEL_FIELD_MAP[name], None)
        # if name == 'programmes':
        #     return self.programmes
        return None


    def __setattr__(self, name: str, value: Any) -> None:
        """Handle attribute assignments."""
        if not name.startswith('_') and name != 'programmes':
            raise AttributeError(f"Can't modify {name}")
        else:
            super().__setattr__(name, value)

    def __repr__(self):
        """Give human-friendly representation."""
        return f'<Channel: sources={self._sources}, id={self.id}, ' + \
        f'xmltv_id={self.xmltv_id}, number={self.number}, name={self.name}>'

    def __hash__(self):
        """Return hash of this object."""
        # both sourced channel
        if (self._sources & CSRC.skyq_service_summary == CSRC.skyq_service_summary) and \
            (self._sources & CSRC.xml_tv == CSRC.xml_tv):
            assert self._chan_dict['t'] == self._chan_dict['xmltv_display_name']
            return hash(self._chan_dict['t'])

        # sky sourced channel
        if self._sources & CSRC.skyq_service_summary == CSRC.skyq_service_summary:
            return hash(self._chan_dict['t'])

        # xmltv sourced channel
        if self._sources & CSRC.xml_tv == CSRC.xml_tv:
            return hash(self._chan_dict['xmltv_display_name'])

        # empty channel
        return hash(CSRC.no_source)

    def __eq__(self, other):
        """Return True if self == other."""
        if not isinstance(other, Channel):
            return False

        # pylint: disable=protected-access
        return self._chan_dict == other._chan_dict and \
            self._sources == other._sources and \
            self.programmes == other.programmes



    def as_json(self) -> str:
        """Return a JSON string respenting this Channel."""
        return json.dumps(self, cls=_ChannelJSONEncoder, indent=4)

    @property
    def sources(self) -> CSRC:
        """Return the sources flag.

        Returns:
            CSRC: A flag of one or more
                :class:`~.constants.CHANNELSOURCES` ORed together to indicate
                which sources have been applied to this object.

        """
        return self._sources # type: ignore

    def load_skyq_summary_data(self,
                               chan_dict: Dict[str, Any]
                               ) -> None:
        """Load summary data from SkyQ box into the Channel.

        Args:
            chan_dict (dict): A detail dict that is obtained from the
                ``as/services/`` endpoint.

        """
        # pylint: disable=attribute-defined-outside-init
        self._chan_dict.update(chan_dict)
        self._sources = (self._sources | CSRC.skyq_service_summary)  # type: ignore

    def load_skyq_detail_data(self, detail_dict: Dict[str, Any]) -> None:
        """Add additional properties obtained from the SkqQ box to the object.

        Args:
            detail_dict (dict): A detail dict that is obtained from the
                ``as/services/details/<sid>`` endpoint.

        """
        # pylint: disable=attribute-defined-outside-init
        self._chan_dict.update(detail_dict['details'])
        self._sources = (self._sources | CSRC.skyq_service_detail)

    def load_xmltv_data(self,
                        xml_chan: Element,
                        base_url: URL = URL('http://www.xmltv.co.uk/'),
                        ) -> None:
        """Take an XML TV element and load it into the channel's data structure.

        This method loads the :class:`~pyskyq.channel.Channel` object with data
        passed in from an XML TV channel element. Specifically, this is either
        additional data about the :class:`~pyskyq.channel.Channel` in question,
        or it is :class:`~pyskyq.programme.Programme` data for this channel.

        Args:
            xml_chan (Element): A XML element of ``<channel>...</channel>`` or
                tags ``<programme>...</programme>``.
            base_url (URL): A :py:class:`~yarl.URL` prefix which is used construct
                the full :attr:`~.channel.Channel.xmltv_icon_url` property.

        """
        assert xml_chan.tag.lower() == 'channel'

        self._chan_dict['xmltv_id'] = xml_chan.attrib['id']
        for child in xml_chan:
            if child.tag.lower() == 'icon':
                self._chan_dict['xmltv_icon_url'] = base_url.join(URL(child.attrib['src']))
                continue
            if child.tag.lower() == 'display-name':
                self._chan_dict['xmltv_display_name'] = child.text

        # pylint: disable=attribute-defined-outside-init
        self._sources = (self._sources | CSRC.xml_tv)

# pylint: disable=attribute-defined-outside-init
# pylint: disable=protected-access
def channel_from_json(json_) -> Channel:
    """Create a channel from JSON data.

    Args:
        str: A string of JSON data.

    Returns:
        Channel: A channel object.

    """
    chan = Channel.__new__(Channel)
    data = json.loads(json_, object_hook=skyq_json_decoder_hook)
    if not data.get('__type__') == '__channel__':
        raise ValueError('Incorrect type metadata in JSON payload.')
    chan._chan_dict = data['attributes']
    chan._sources = data['sources']
    chan.programmes = SortedSet()
    for prog in data['programmes']:
        chan.programmes.add(
            Programme(channel_xml_id=data['attributes']['xmltv_id'], **prog['attributes'])
        )
    return chan



def channel_from_skyq_service(skyq_chan: Dict[str, Any]) -> Channel:
    """Create a Channel object from a SkyQ Service summary payload.

    Args:
        skyq_chan (Dict[str, Any]): This dictionary is the payload that comes directly from the
            SkyQ's ``as/services/`` endpoint.

    Returns:
        Channel: A channel object with the  SkyQ summary data loaded.

    """
    chan = Channel.__new__(Channel)
    chan.load_skyq_summary_data(skyq_chan)
    return chan

def channel_from_xmltv_list(xml_chan: Element) -> Channel:
    """Create a Channel object from an XMLTV channel element.

    This function is a Channel factory. It will generate a Channel object given a XML
    channel element from an XML TV file.

    Args:
        xml_chan (Element): A XML element of ``<channel>...</channel>`` tags.

    Returns:
        Channel: A channel object with the XML TV data loaded.

    """
    chan = Channel.__new__(Channel)
    chan.load_xmltv_data(xml_chan)
    return chan

def merge_channels(chan_a: Channel,
                   chan_b: Channel,
                   ) -> Channel:
    """Merge two channels together and returns a new one with the data combined.

    Args:
        chan_a (Channel): First channel to merge.
        chan_b (Channel): Second channel to merge.

    Returns:
        Channel: Merged channel, although see warning for gotchas.

    Warning:
        No checks are made to ensure these two channels represent the same real channel.
        You should check each first before merging, perhaps by matching on the ``name``
        or ``xmltv_display_name``.

    If there are any key clashes ``chan_b``'s keys will overrite ``chan_a``'s with
    the exception of the following:

    - if the ``sid`` or ``id`` attribute is ``None`` on one and not the other, the other
        will override regardless of the passed order.
    - if the ``c`` or ``number`` attribute is ``None`` on one and not the other, the other
        will override regardless of the passed order.
    - if the ``t`` or ``name`` attribute is ``None`` on one and not the other, the other
        will override regardless of the passed order.
    - if the ``xmltv_id`` attribute is ``None`` one one and not the other, the other will
        override, regardless of the passed order.

    """
    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=protected-access

    new_chan = Channel.__new__(Channel)

    new_chan._chan_dict = {**chan_a._chan_dict, **chan_b._chan_dict}

    if chan_a.id is not None and chan_b.id is None:
        new_chan._chan_dict['sid'] = chan_a.id

    if chan_b.id is not None and chan_a.id is None:
        new_chan._chan_dict['sid'] = chan_b.id


    if chan_a.c is not None and chan_b.c is None:
        new_chan._chan_dict['c'] = chan_a.c

    if chan_b.c is not None and chan_a.c is None:
        new_chan._chan_dict['c'] = chan_b.c


    if chan_a.t is not None and chan_b.t is None:
        new_chan._chan_dict['t'] = chan_a.t

    if chan_b.t is not None and chan_a.t is None:
        new_chan._chan_dict['t'] = chan_b.t


    if chan_a.xmltv_id is not None and chan_b.xmltv_id is None:
        new_chan._chan_dict['xmltv_id'] = chan_a.xmltv_id

    if chan_b.xmltv_id is not None and chan_a.xmltv_id is None:
        new_chan._chan_dict['xmltv_id'] = chan_b.xmltv_id

    new_chan._sources = chan_a._sources | chan_b._sources
    new_chan.programmes = chan_b.programmes.union(chan_a.programmes)
    return new_chan



import logging
import sys
import trio
import pytest
from trio_websocket import ConnectionClosed, WebSocketConnection
from contextlib import asynccontextmanager
from functools import partial
from asynctest import CoroutineMock, MagicMock
from pyskyq import get_status
from .http_server import http_server, websocket_server

from .asynccontextmanagermock import AsyncContextManagerMock
from .mock_constants import WS_STATUS_MOCK

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.WARNING, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

LOGGER = logging.getLogger(__name__)


async def test_get_status():

    with trio.move_on_after(1):
        async with trio.open_nursery() as server_nursery:
            responses = []
            response = {
                'target': '/as/system/status',
                'body': WS_STATUS_MOCK.encode('utf-8'),
                }

            for _ in range(10):
                responses.append(response)  # add 10 of them for good measure.
            server_nursery.start_soon(websocket_server, '127.0.0.1', 9006, responses)
            # await trio.serve_tcp(partial(http_server, responses=responses), 8000)
            LOGGER.debug('started websocket server.')
            await trio.sleep(0.3)
            # pylint: disable=not-async-context-manager

            async with get_status('localhost') as stat:
                await trio.sleep(0.1)
                assert stat.online is True  #object initialises this to False, JSON payload sets to True.




# pylint: skip-file

import contextlib
import os
import shutil
import tempfile

# Shamelessly stolen from the excellent Click library:
# See https://github.com/pallets/click/blob/master/click/testing.py

@contextlib.contextmanager
def isolated_filesystem():
    """A context manager that creates a temporary folder and changes
    the current working directory to it for isolated filesystem tests.
    """
    cwd = os.getcwd()
    t = tempfile.mkdtemp()
    os.chdir(t)
    try:
        yield t
    finally:
        os.chdir(cwd)
        try:
            shutil.rmtree(t)
        except (OSError, IOError):
            pass
